package com.cotopia.games;


import com.cotopia.games.IllegalPositionException;
import java.util.Enumeration;

import javax.microedition.lcdui.Canvas;
import javax.microedition.lcdui.Choice;
import javax.microedition.lcdui.Command;
import javax.microedition.lcdui.CommandListener;
import javax.microedition.lcdui.Display;
import javax.microedition.lcdui.Displayable;
import javax.microedition.lcdui.Font;
import javax.microedition.lcdui.Graphics;
import javax.microedition.lcdui.Image;
import javax.microedition.lcdui.List;
import com.cotopia.games.Play;
import com.cotopia.games.BallShotRules;

/**
 * @author karel herink
 */
//#if NokiaUI
public class BallShotBoard extends com.nokia.mid.ui.FullCanvas implements CommandListener, Runnable {
//#else
//# public class BallShotBoard extends javax.microedition.lcdui.Canvas implements CommandListener, Runnable {
//#endif

	private byte level = 0;
	
	//if "true" gun moves on it's own
	private boolean autoMode;
			
	private final static long REPAINT_PERIOD_MS = 40;
	private final static long FLASH_PERIOD_MS = 1000;
	private long startFlashModifiedBoardTime;
	
	
	public byte SCREEN_SINGLE_PLAYER_IN_GAME_MENU = 120;
	private byte CANON_ROWS = 2;
	
	private byte STATUS_BAR_UNIT = 20;
	private boolean statusVertical;
	private short statusY;
	private short statusXCenterVerIcons;
	
	boolean running;
	private byte[][] board;
	
	private BallShotRules rules;
	private int anchor;
	
	Gun canon;

	//playing board
	short leftXBoundry;
	short rightXBoundry;
	short topYBoundry;
	short bottomYboundry;

	short ballsOffsetX;
	short ballsOffsetY;
	short ballsWidth;
	short ballsHeight;
	short[] ballsTopLeft;
	short[] ballsTopRight;
	
	private short ballsTotalLeftShortRowOffset;

	short canonTop;
	short canonLeft;
	short canonHeight;
	short canonWidth;
	
	short width;
	short height;

	byte fieldEdgeLenght;
	
	private Ball ball;
	
	private static final int BALL_FIRE_LIMIT_MS = 10000;

	private long ballLastFiredTime;
	private long rowInsertLimitMS;
	private long rowLastInsertTime;
	private long lostTimeMS;

	private boolean rowInsertWarning;

	private BallShot midlet;
	
	public final static boolean DEBUG = false;
	public final static byte UNKNOWN = -1;

	protected Display display;
	
	protected BallShotImageCreator imageCreator;
	
	public BallShotBoard(BallShot midlet, BallShotRules rules, int rowInsertLimitMS) {
		this.display = Display.getDisplay(midlet);
		this.midlet = midlet;
		this.imageCreator = new BallShotImageCreator();

		this.rowInsertLimitMS = rowInsertLimitMS;
		
		this.rules = rules;
		this.anchor = Graphics.TOP | Graphics.LEFT;
		
		this.rules.initLevel(++this.level);
		
		this.width = (short) this.getWidth();
		this.height = (short) this.getHeight();

		if (DEBUG)
			System.out.println("Width = " + width + ", Height = " + height);
		
		//determine where the status bar will be painted
		if (this.width > this.height) {
			this.width -= STATUS_BAR_UNIT;
			this.statusVertical = true;
			if (DEBUG)
				System.out.println("Vertical");
		}
		else {
			this.height -= STATUS_BAR_UNIT;
			if (DEBUG)
				System.out.println("Horizontal");
		}

		if (DEBUG)
			System.out.println("Width = " + width + ", Height = " + height);
		
		
		this.fieldEdgeLenght = this.imageCreator.getBallDiameter();
		//make sure we work with even numbers 
		this.fieldEdgeLenght = (byte) ((this.fieldEdgeLenght % 2) != 0 ? this.fieldEdgeLenght -1 : this.fieldEdgeLenght);
		if (DEBUG)
			System.out.println("Edge = " + this.fieldEdgeLenght);
		
		
		
		this.ballsOffsetX = (short) ((this.width - (fieldEdgeLenght * BallShotRules.FULL_COLS)) / 2);
		this.ballsOffsetY = (short) ((this.height - (fieldEdgeLenght * (BallShotRules.BOARD_ROWS + CANON_ROWS))) / 2);
		if (DEBUG)
			System.out.println("ballsOffsetX = " + ballsOffsetX + " , ballsOffsetY = " + ballsOffsetY);
		
		this.ballsTotalLeftShortRowOffset = (short) (this.fieldEdgeLenght / 2 + this.ballsOffsetX);
		this.ballsHeight = (short) (fieldEdgeLenght * (BallShotRules.BOARD_ROWS));
		this.ballsWidth = (short) (fieldEdgeLenght * BallShotRules.FULL_COLS);

		this.canonTop = (short) (fieldEdgeLenght * BallShotRules.BOARD_ROWS + this.ballsOffsetY);
		this.canonLeft = ballsOffsetX;
		this.canonHeight = (short) (fieldEdgeLenght * CANON_ROWS);
		this.canonWidth = ballsWidth;
		
		this.leftXBoundry = ballsOffsetX;
		this.rightXBoundry = (short) (ballsOffsetX + ballsWidth);
		this.topYBoundry = this.ballsOffsetY;
		this.bottomYboundry = this.canonTop;
		
		this.ballsTopLeft = new short[] {this.leftXBoundry, this.topYBoundry};
		this.ballsTopRight = new short[] {this.rightXBoundry, this.topYBoundry};
		
		this.statusY = (short) (this.canonTop + canonHeight + 1);
		this.statusXCenterVerIcons = (short) (this.rightXBoundry + (this.getWidth() - this.rightXBoundry) / 2);
		
		
		this.canon = new Gun(this, this.fieldEdgeLenght, rowInsertLimitMS);
		this.running = true;
		this.setAutoMode(true);
	}

	public void run() {
		if (DEBUG)
			System.out.println("BallShotBoard thread RUNNING... ");
		
		this.rowLastInsertTime = this.ballLastFiredTime = System.currentTimeMillis();
		
		while (running) {
			long start = System.currentTimeMillis();
			
			//this allows for displaying the playing board for a few seconds before showing the menu
			if ( (this.lostTimeMS != 0) && (System.currentTimeMillis() - this.lostTimeMS >= 4000) ) {
				//this.drawInfoTextBox(null, anchor, getFieldH(), getFieldW())
				if (DEBUG)
					System.out.println("LOSE REPAINT FINISHED");
				this.running = false;
				break;
			}
						
			if (this.startFlashModifiedBoardTime == 0) {
				this.board = rules.getBoard();
			}
			else {
				if ((this.startFlashModifiedBoardTime + FLASH_PERIOD_MS) < start) {
					if (DEBUG)
						System.out.println("stop flashing");
					this.board = rules.getBoard();
					this.startFlashModifiedBoardTime = 0;
					if (this.autoMode)
						this.canon.setPaused(false);
				}
			}
			
			this.repaint();
			this.serviceRepaints();
			try {
				long diff = start + REPAINT_PERIOD_MS - System.currentTimeMillis();
				if ( diff > 0)
					Thread.sleep(diff);
			}
			catch (InterruptedException e) {
				if (DEBUG)
					e.printStackTrace();
			}
			
			//shoot if needed
			long now = System.currentTimeMillis();
			if (BALL_FIRE_LIMIT_MS < now - this.ballLastFiredTime) {
				this.canon.shoot();
				this.ballLastFiredTime = now;
			}
			
				if (rowInsertLimitMS < now - rowLastInsertTime) {
					if (DEBUG)
						System.out.println("rowInsertLimitMS < now - rowLastInsertTime");
					this.rowInsertWarning = false;
					byte status  = this.rules.insertRow();
					if (DEBUG)
						System.out.println("this.rules.insertRow()");
					this.rowLastInsertTime = now;
					if (status == BallShotRules.LOSE) {
						this.lostTimeMS = System.currentTimeMillis();
						this.ball = null;
						this.canon.setPaused(true);
					}
				}
				else if (rowInsertLimitMS - 5000 < now - rowLastInsertTime) {
					this.rowInsertWarning = true;
				}
		}
		if (DEBUG)
			System.out.println("\n\t--------------\n\tBallShotBoard STOPPED RUNNING");
		this.cleanReferencesForExit();
		this.exitGame();
	}
	
	public void exitGame() {
		this.midlet.showMenu();
	}
	
	public void paint(Graphics g) {
		this.tileBackground(g);

		g.setClip(this.ballsOffsetX, this.ballsOffsetY, this.ballsWidth, this.ballsHeight);
		g.setColor(0xddffcc);
		g.fillRect(0, 0, this.getWidth(), this.getHeight());
		
		g.setClip(0, 0, this.getWidth(), this.getHeight());
		
		//then paint the red row
		g.setColor(0xffa1a6);
		g.fillRect(ballsOffsetX+1, ballsOffsetY + (this.fieldEdgeLenght * (BallShotRules.BOARD_ROWS - 1)), this.ballsWidth-1, this.fieldEdgeLenght);

		this.paintBalls(g);
		this.canon.paint(g);
		if (this.statusVertical)
			this.paintStatusBarVer(g);
		else
			this.paintStatusBarHor(g);
		if (this.ball != null)
			ball.paint(g);		
	}

	private void paintStatusBarVer(Graphics g) {
		g.setFont(Font.getFont(Font.FACE_MONOSPACE, Font.STYLE_BOLD, Font.SIZE_LARGE));
		g.setColor(0xffffff);
		
		int anchor = Graphics.TOP | Graphics.HCENTER;
		short iconYOff = (short) ((this.ballsHeight + this.canonHeight) / 5);
		short y = this.ballsOffsetY;
		if (this.rules.remoteInProgress || this.rowInsertWarning) {
			g.drawImage(((BallShotImageCreator) imageCreator).getIcon(BallShotImageCreator.WARN), this.statusXCenterVerIcons, y, anchor);
		}
		y += iconYOff;
		y += iconYOff;
		g.drawImage((Image) ((BallShotImageCreator) imageCreator).getImage(rules.getNextBall(), UNKNOWN), this.statusXCenterVerIcons, y, Graphics.TOP | Graphics.HCENTER);
		y += iconYOff;
		g.setColor(0x00AA00);
		g.drawString(Integer.toString(this.rules.getNumBallsSent()), this.statusXCenterVerIcons, y, anchor);
		y += iconYOff;
		g.setColor(0xAA0000);
		g.drawString(Integer.toString(this.rules.getNumBallsRecieved()), this.statusXCenterVerIcons, y, anchor);
	}
	
	private void paintStatusBarHor(Graphics g) {
		g.setFont(Font.getFont(Font.FACE_MONOSPACE, Font.STYLE_BOLD, Font.SIZE_LARGE));
		g.setColor(0xffffff);
		
		int anchor = Graphics.TOP | Graphics.HCENTER;
		short iconXOff = (short) (this.ballsWidth / 5);
		short x = (short) (this.ballsOffsetX + iconXOff / 2);
		if (this.rules.remoteInProgress || this.rowInsertWarning) {
			g.drawImage(((BallShotImageCreator) imageCreator).getIcon(BallShotImageCreator.WARN), x, this.statusY, anchor);
		}
		x += iconXOff;
		g.setColor(0xAA0000);
		g.drawString(Integer.toString(this.rules.getNumBallsRecieved()), x, this.statusY, anchor);
		x += iconXOff;
		g.drawImage((Image) ((BallShotImageCreator) imageCreator).getImage(rules.getNextBall(), UNKNOWN), x, this.statusY, anchor);
		x += iconXOff;
		g.setColor(0x00AA00);
		g.drawString(Integer.toString(this.rules.getNumBallsSent()), x, this.statusY, anchor);
		x += iconXOff;
	}
	
	private void tileBackground(Graphics g) {
		Image tile = ((BallShotImageCreator) imageCreator).getIcon(BallShotImageCreator.TILE);
		//tile the whole screen firsts
		if (tile == null) {
			//blue screen
			g.setColor(0xffffff);
			g.fillRect(0, 0, this.width, this.height);
		}
		else {
			byte tileRows = (byte) (this.getHeight() / BallShotImageCreator.TILE_HEIGHT);
			byte tileCols = (byte) (this.getWidth() / BallShotImageCreator.TILE_WIDTH);
			for (int r = 0; r <= tileRows; r++) {
				for (int c = 0; c <= tileCols; c++) {
					g.drawImage(tile, c * BallShotImageCreator.TILE_WIDTH, r * BallShotImageCreator.TILE_HEIGHT, Graphics.TOP | Graphics.LEFT);
				}
			}
		}
	}
	
	private void paintBalls(Graphics g) {
		int y = ballsOffsetY;

		g.setColor(0x000000);
		g.drawRect(this.ballsOffsetX, y, this.ballsWidth, this.ballsHeight);
		for (byte i = 0; i < this.board.length; i++) {
			
			y = (i * this.fieldEdgeLenght) + ballsOffsetY;
			
			//if this is a short row paint as such
			if (board[i].length < BallShotRules.FULL_COLS) {
				for (byte j = 0; j < board[i].length; j++)
					if (board[i][j] != BallShotRules.NO_PIECE)
						g.drawImage((Image) ((BallShotImageCreator) imageCreator).getImage(board[i][j], UNKNOWN), j * this.fieldEdgeLenght +  ballsTotalLeftShortRowOffset, y, anchor);
			}
			//else it must be a regular long row
			else {
				for (byte j = 0; j < board[i].length; j++)
					if (board[i][j] != BallShotRules.NO_PIECE)
						g.drawImage((Image) ((BallShotImageCreator) imageCreator).getImage(board[i][j], UNKNOWN), j * this.fieldEdgeLenght + this.ballsOffsetX, y, anchor);
			}
		}
	}

	
	/////////////////////////////////////////////////////////////////////////////////////////
	// NEEDED FOR BALL COLLISIONS AND PLACEMENT
	/////////////////////////////////////////////////////////////////////////////////////////
	
	
	public void setBall(byte type, int x, int y) {
		byte status = BallShotRules.VALID_UNFINISHED;
		byte[] field = getFieldforPosition(x, y);
		
		try {
			status = this.rules.executeMove(BallShotRules.LOCAL_MOVE, type, field[0], field[1]);
		} catch (IllegalPositionException e) {
			if (DEBUG)
				e.printStackTrace();
		}
		
		if (status == BallShotRules.VALID_FINISHED && ((Play)this.rules.getPlay()).getAllMoves().size() > 0) {
			if (DEBUG)
				System.out.println("start flashing");
			//let everyone know that for a while we'll display modified board
			this.startFlashModifiedBoardTime = System.currentTimeMillis();
			//copy original board from the rules
			this.board = new byte[BallShotRules.BOARD_ROWS][];
			for (byte i = 0; i < board.length; i++) {
				board[i] = new byte[this.rules.getBoard()[i].length];
				for (byte j = 0; j < board[i].length; j++) {
					this.board[i][j] = this.rules.getBoard()[i][j];
				}
			}
			//add deleted balls from the play to the local board
			Enumeration e = ((Play)this.rules.getPlay()).getAllMoves().elements();
			byte[] move;
			while (e.hasMoreElements()) {
				move = (byte[]) e.nextElement();
				this.board[move[2]][move[3]] = BallShotRules.BALL_DROPPED;
			}
			((Play)this.rules.getPlay()).clearMoves();			
			this.ball = null;
		}
		else if (status == BallShotRules.LOSE) {
			this.running = false;
			try {
				Thread.sleep(2000);
			}
			catch (InterruptedException e) {
				e.printStackTrace();
			}
		}
		else {
			this.ball = null;
			if (this.autoMode)
				this.canon.setPaused(false);
		}
	}
	
	public boolean isBallCollisionAtPosition(int posX, int posY) {
		byte[] field = this.getFieldforPosition(posX, posY);
		if (DEBUG)
			System.out.println("\t\tROW = " + field[0] + "\tCOL = " + field[1]);
		byte ball = BallShotRules.NO_PIECE;
		try {
			ball = rules.getPieceAtPosition(field[0], field[1]);
		} catch (IllegalPositionException e) {
			if (DEBUG)
				e.printStackTrace();
		}
		if (ball != BallShotRules.NO_PIECE) {
			return true;
		}
		return false;
	}
	
	public byte[] getFieldforPosition(int posX, int posY) {
		byte r = (byte) ( (posY - this.ballsOffsetY) / this.fieldEdgeLenght );
		
		r = r < 0 ? 0 : r;
		r = r >= BallShotRules.BOARD_ROWS ? BallShotRules.BOARD_ROWS -1 : r;
		

		if (DEBUG)
			System.out.println("ROW = " + r);
		byte c;
		
		if (rules.getBoard()[r].length == BallShotRules.SHORT_COLS) {
			if (DEBUG)
				System.out.println("SHORT ROW");
			c = (byte) ((posX - this.ballsTotalLeftShortRowOffset) / this.fieldEdgeLenght);
			//if the ball went out of X bounds first, we maye ge a number 1 bigger here 
			c = c >= BallShotRules.SHORT_COLS-1 ? BallShotRules.SHORT_COLS-1 : c;
		}
		else {
			if (DEBUG)
				System.out.println("LONG ROW");
			c = (byte) ((posX - this.ballsOffsetX) / this.fieldEdgeLenght);
			c = c >= BallShotRules.FULL_COLS-1 ? BallShotRules.FULL_COLS-1 : c;
		}
		c = c < 0 ? 0 : c;
		return new byte[] {r, c};
	}
	
	
	/////////////////////////////////////////////////////////////////////////////////////////
	// HELPERS
	/////////////////////////////////////////////////////////////////////////////////////////

	//all the fields must be sqaure so edge is assigned the smalled value of potential 
	//field width and potential field height
	private byte calculateFieldEdge(byte rows, byte maxCols) {
		byte h = (byte) (this.height / rows);
		byte w = (byte) (this.width / maxCols);
		return (h < w ? h : w);
	}

	public void setBall(Ball ball) {
		this.ball = ball;
	}
	public Ball getBall() {
		return this.ball;
	}

	public BallShotRules getRules() {
		return this.rules;
	}	
	
	protected void keyReleased(int arg0) {
		if (!this.autoMode)
			this.canon.setPaused(true);
	}
	
	private void keyPressedAuto(int code) {
		if (DEBUG)
			System.out.println("keyPressedAuto");
			
		switch (this.getGameAction(code)) {
			case Canvas.UP :
				this.canon.faster();
				if (DEBUG)
					System.out.println("Cannon period : " + this.canon.canonPeriodMS);
				break;
			case Canvas.DOWN :
				this.canon.slower();
				if (DEBUG)
					System.out.println("Cannon period : " + this.canon.canonPeriodMS);
				break;
			case Canvas.RIGHT :
				if (DEBUG)
					System.out.println("Cannon direction : RIGHT");
				this.canon.setRight();
				break;
			case Canvas.LEFT :
				if (DEBUG)
					System.out.println("Cannon direction : LEFT");
				this.canon.setLeft();
				break;
				
			default :
				keyPressedCommon(code);
				break;
		}
	}
	
	private void keyPressedManual(int code) {
		if (DEBUG)
			System.out.println("keyPressedManual");
		switch (this.getGameAction(code)) {
			case Canvas.UP :
				this.canon.faster();
				if (DEBUG)
					System.out.println("Cannon period : " + this.canon.canonPeriodMS);
				break;
			case Canvas.DOWN :
				this.canon.slower();
				if (DEBUG)
					System.out.println("Cannon period : " + this.canon.canonPeriodMS);
				break;
			case Canvas.RIGHT :
				if (DEBUG)
					System.out.println("Cannon direction : RIGHT");
				this.canon.setRight();
				//if (this.ball == null)
				this.canon.setPaused(false);
				break;
			case Canvas.LEFT :
				if (DEBUG)
					System.out.println("Cannon direction : LEFT");
				this.canon.setLeft();
				//if (this.ball == null)
				this.canon.setPaused(false);
				break;
				
			default :
				keyPressedCommon(code);
				break;
		}
	}
	
	private void keyPressedCommon(int code) {
		if (DEBUG)
			System.out.println("keyPressedCommon");
		switch (this.getGameAction(code)) {
			case Canvas.FIRE :
				if (this.lostTimeMS == 0) {
					if (DEBUG)
						System.out.println("SHOOTING");
					this.canon.shoot();
					this.ballLastFiredTime = System.currentTimeMillis();
				}
				break;
			case Canvas.GAME_C :
				//fallthrough
			case Canvas.GAME_D :
				//only option to quit or go back
				showSinglePlayerInGameMenu();
				break;
				
			default :
					break;
		}
	}
	
	protected void keyPressed(int code) {
		if (this.autoMode) {
			this.keyPressedAuto(code);
		}
		else {
			this.keyPressedManual(code);
		}
	}

	private void showSinglePlayerInGameMenu() {
		if (DEBUG)
			System.out.println("showSinglePlayerInGameMenu");
		List inGameMenu = new List(BallShotRules.MSG_CHOOSE_OPTION, Choice.IMPLICIT, new String[] {BallShotRules.MSG_BACK, autoMode ? BallShotRules.MSG_SET_MANUAL_CNTRL : BallShotRules.MSG_SET_AUTO_CNTRL, BallShotRules.MSG_QUIT_GAME}, null);
		inGameMenu.setCommandListener(this);
		display.setCurrent(inGameMenu);
	}
	
	public void commandAction (Command c, Displayable d) {
		if (DEBUG) {
			System.out.println("handling SCREEN_SINGLE_PLAYER_IN_GAME_MENU");
		}
		int index = ((List) d).getSelectedIndex();
		//if "BACK" selected then show board again
		if (index == 0) {
			this.showBoard();
		}
		//switch the auto mode
		else if (index == 1) {
			this.setAutoMode(!this.autoMode);
			this.showBoard();
		}
		//or "QUIT" was selected 
		else {
			this.running = false;
		}
	}	
	
	public void cleanReferencesForExit() {
		if (DEBUG)
			System.out.println("Board calling cleanReferencesForExit()");
		this.imageCreator.cleanReferences();
		this.imageCreator = null;
		System.gc();
	}
	
	
	public void setImageCreator(BallShotImageCreator imageCreator) {
		this.imageCreator = imageCreator;
	}

	public void handleMoveExecuted(byte moveStatus, boolean isBoardChanged) {
		repaintPlayingArea();
	}
		
	public void repaintPlayingArea() {
		this.repaint(0, 0, this.getWidth(), this.getHeight());
	}

	public BallShotImageCreator getImageCreator() {
		return imageCreator;
	}


	public void showBoard() {
		display.setCurrent(this);
	}
	public boolean isRunning() {
		return running;
	}

	public void setRunning(boolean running) {
		this.running = running;
	}
	/**
	 * @return Returns the autoMode.
	 */
	public boolean isAutoMode() {
		return this.autoMode;
	}

	/**
	 * @param autoMode The autoMode to set.
	 */
	public void setAutoMode(boolean autoMode) {
		this.canon.setPaused(!autoMode);
		this.autoMode = autoMode;
	}

}
