package com.cotopia.games;

import java.io.IOException;
import java.util.Random;

import javax.microedition.lcdui.Graphics;
import javax.microedition.lcdui.Image;

import com.cotopia.games.BallShotRules;

/**
 * @author karel herink
 */
public class BallShotImageCreator {

	public static final boolean DEBUG = false;
	
	public static final byte WARN = 126;
	public static final byte GUN_FOREGROUND = 127;
	public static final byte TILE = 124;
	public static final byte TILE_WIDTH = 20;
	public static final byte TILE_HEIGHT = 20;
	
	private Image tile;
	private Image gunForeground;
	private Image warn;
	
	
	private Image[] images = new Image[BallShotRules.NUM_BALL_TYPES];

	private Random rand = new Random();

	private byte ballWidth;
	
	
	public BallShotImageCreator() {
		if (DEBUG) {
			System.out.println("BallShotImageCreator()");
		}
		this.ballWidth = 16;
		
//#if screen == "128x128"
		ballWidth = 10;
//#elif screen == "176x182"
//# 		ballWidth = 16;
//#elif screen == "176x208"
//# 		ballWidth = 16;
//#endif
		
		this.images[0] = createColorBallImage(BallShotRules.BALL_BLACK, ballWidth);
		this.images[1] = createColorBallImage(BallShotRules.BALL_RED, ballWidth);
		this.images[2] = createColorBallImage(BallShotRules.BALL_BLUE, ballWidth);
		this.images[3] = createColorBallImage(BallShotRules.BALL_GREEN, ballWidth);
		this.images[4] = createColorBallImage(BallShotRules.BALL_WHITE, ballWidth);
		this.images[5] = createColorBallImage(BallShotRules.BALL_PINK, ballWidth);
		this.images[6] = createColorBallImage(BallShotRules.BALL_YELLOW, ballWidth);
		try {
			this.tile = loadIcon(TILE, ballWidth);
			this.warn = loadIcon(WARN, ballWidth);
			this.gunForeground = loadIcon(GUN_FOREGROUND, ballWidth);
		} catch (IOException e) {
			if (DEBUG)
				e.printStackTrace();
		}
	}

	public byte getBallDiameter() {
		return this.ballWidth;
	}
	
	private Image createColorBallImage(byte ballColor, int ballWidth) {
		Image img = null;
		try {
			String imgStr = "/res/" + ballWidth + "_" + ballColor + ".png";
			if (DEBUG)
				System.out.println("Creating image " + imgStr);
			img = Image.createImage(imgStr);
		}
		catch (IOException e) {
			e.printStackTrace();
		}

		if (img != null)
			return img;
		
		img = Image.createImage(ballWidth, ballWidth);
		Graphics g = img.getGraphics();
		g.setColor(0xddffcc);
		g.fillRect(0, 0, ballWidth, ballWidth);
		int outline = 0xFFFFFF;
		int fillColor = 0;
		switch (ballColor) {
			case BallShotRules.BALL_BLACK :
				fillColor = 0xAA00AA;
				break;
			case BallShotRules.BALL_RED :
				fillColor = 0xFF0033;
				break;
			case BallShotRules.BALL_BLUE :
				fillColor = 0x00bbFF;
				break;
			case BallShotRules.BALL_GREEN :
				fillColor = 0x00EE00;
				break;
			case BallShotRules.BALL_WHITE :
				fillColor = 0xFFCC00;
				break;
			case BallShotRules.BALL_PINK :
				fillColor = 0xFF33FF;
				break;
			case BallShotRules.BALL_YELLOW:
				fillColor = 0x000000;
				break;
		}
		g.setColor(fillColor);
		g.fillArc(1, 1, ballWidth-2, ballWidth-2, 0, 360);
		g.setColor(outline);
		g.drawArc(0, 0, ballWidth-1, ballWidth-1, 0, 360);
		return img;
	}

	private Image loadIcon(byte type, byte ballDiameter) throws IOException {
		Image img = null;
		
		switch (type) {
			case TILE :
				img = Image.createImage("/res/TILE.png");
				break;
			case GUN_FOREGROUND :				
				img = Image.createImage("/res/"+ ballDiameter +"_cb.png");
				break;
			case WARN :
				img = Image.createImage("/res/WARN.png");
				break;
			default :
				break;
		}
		return img;
	}
	
	public Object getImage(int pieceType, int bgColor) {
		byte ball = (byte) pieceType;
		byte ballColor = (byte) (ball & BallShotRules.COLOR_MASK);
		switch (ballColor) {
			case BallShotRules.BALL_BLACK :
				return images[0];
			case BallShotRules.BALL_RED :
				return images[1];
			case BallShotRules.BALL_BLUE :
				return images[2];
			case BallShotRules.BALL_GREEN :
				return images[3];
			case BallShotRules.BALL_WHITE :
				return images[4];
			case BallShotRules.BALL_PINK :
				return images[5];
			case BallShotRules.BALL_YELLOW :
				return images[6];
			case BallShotRules.BALL_DROPPED :
				return getRandomImage();
			default :
				if (DEBUG)
					System.out.println("Error asking BallShotImageCreator for image for unknown pieceType: " + ballColor);
				return null;
		}
	}
	
	private Image getRandomImage() {
		return this.images[Math.abs(this.rand.nextInt() % BallShotRules.NUM_BALL_TYPES)];
	}
	
	
	public Image getIcon(int type) {
		switch (type) {
			case BallShotImageCreator.TILE :
				return tile;
			case BallShotImageCreator.GUN_FOREGROUND :
				return gunForeground;
			case BallShotImageCreator.WARN :
				return warn;
			default :
				if (DEBUG)
					System.out.println("Unknown ICON: " + type);
				return null;
			}
	}

	/* (non-Javadoc)
	 * @see com.digitalschemes.boardgames.common.ImageCreator#cleanReferences()
	 */
	public void cleanReferences() {
		this.images = null;
		this.tile = null;
		this.gunForeground = null;
		this.warn = null;
	}

}
