package com.cotopia.games;

import java.util.Random;

/**
 * @author karel herink
 */
public class BallShotRules {
	public final static byte NO_PIECE = 0;
 	public static final byte VALID_UNFINISHED = 0;
	public static final byte VALID_FINISHED = 1;
	public static final byte INVALID = 2;
	public static final byte LOSE = 4;

	public final static String MSG_CHOOSE_OPTION	= "Choose option:";
	public final static String MSG_BACK				= "Back";
	public static final String MSG_SET_MANUAL_CNTRL = "Set manual controls";
	public static final String MSG_SET_AUTO_CNTRL 	= "Set automatic controls";
	public static final String MSG_QUIT_GAME 		= "Quit game";
	
	
	
	public static final int PERIOD_EASY = 60000;
	public static final int PERIOD_MEDIUM = 50000;
	public static final int PERIOD_HARD = 40000;
	
	//debuging variables
	private final static boolean chains = false;
	private final static boolean DEBUG = false;
	
	
	public static final byte NUM_BALL_TYPES = 7;
	
	public static final byte BALL_BLACK  = 0x01;
	public static final byte BALL_RED 	 = 0x02;
	public static final byte BALL_BLUE 	 = 0x03;
	public static final byte BALL_GREEN	 = 0x04;
	public static final byte BALL_WHITE	 = 0x05;
	public static final byte BALL_PINK	 = 0x06;
	public static final byte BALL_YELLOW = 0x07;
	
	public static final byte BALL_DROPPED	= 0x0E;
	
	
	
	public static final byte COLOR_MASK		= 0x0F;
	public static final byte MARKER_MASK	= 0x10;
	
	public static final byte REMOTE_MOVE	= 0;
	public static final byte LOCAL_MOVE		= 1;

	public static final byte BOARD_ROWS = 8;
	public static final byte FULL_COLS = 10;
	public static final byte SHORT_COLS = FULL_COLS - 1;
	
	//jagged array of alternating SHORT_COLS and FULL_COLS
	private byte[][] board = new byte[BOARD_ROWS][];
	private Play currentPlay;
	
	private byte numBallsMarked = 0;
	
	private Random random = new Random(System.currentTimeMillis());
	private byte currentBall = this.getRandomBall();
	private byte nextBall = this.getRandomBall();
	
	private short numBallsOnBoard = 0;
	
	private short numBallsRecieved = 0;
	private short numBallsSent = 0;

	public boolean remoteInProgress;
	
	public BallShotRules() {
		currentPlay = new Play();
		initBoard();
	}
	
	//initialize the jagged 2D array
	private void initBoard() {
		for (byte i = 0; i < board.length; i++) {
				board[i] = ( i % 2 == 0) ? new byte[SHORT_COLS] : new byte[FULL_COLS];
		}
	}

	
	public void clearBoard() {
		for (byte i = 0; i < board.length; i++) {
			for (byte j = 0; j < board[i].length; j++) {
				board[i][j] = NO_PIECE;
			}
		}
	}
	
	public void initLevel(int level) {
		//level is a row
		for (byte i = 0; i < level; i ++) {
			for (byte col = 0; col < this.board[i].length; col++) {
				board[i][col] = this.determineUniqueBall(i, col);
			}
		}
	}
	
	public byte insertRow() {
		if (DEBUG)
			System.out.println("init the new board");
		byte[][] newBoard = new byte[BallShotRules.BOARD_ROWS][];
		if (this.board[0].length == BallShotRules.SHORT_COLS) {
			for (byte i = 0; i < board.length; i++) {
				newBoard[i] = ( i % 2 == 0) ? new byte[FULL_COLS] : new byte[SHORT_COLS];
			}
		}
		else {
			for (byte i = 0; i < board.length; i++) {
				newBoard[i] = ( i % 2 == 0) ? new byte[SHORT_COLS] : new byte[FULL_COLS];
			}
		}
		
		if (DEBUG)
			System.out.println("copy the existing balls");
		for (byte i = 0; i < board.length -1; i++) {
			for (byte j = 0; j < board[i].length; j++) {
				newBoard[i+1][j] = board[i][j];
			}
		}
		
		if (DEBUG)
			System.out.println("insert the initial row");
		for (byte c = 0; c < newBoard[0].length; c+=2) {
			if (DEBUG)
				System.out.println("c = " + c);
			newBoard[0][c] = this.determineUniqueBall((byte) 0, c);
			this.numBallsRecieved++;
		}

		if (DEBUG)
			System.out.println("assign the new board & clean the old board");
		this.board = newBoard;
		System.gc();
		
		if (DEBUG)
			System.out.println("check if got any balls got pushed below the red line");
		for (byte b = 0; b < this.board[BOARD_ROWS-1].length; b++) {
			if (this.board[BOARD_ROWS-1][b] != NO_PIECE) {
				return LOSE;
			}
		}
		return VALID_UNFINISHED;
	}
	
	public synchronized byte executeMove(byte moveOrigin, byte ball, byte rowTo, byte colTo) 
	throws IllegalPositionException {
		byte moveStatus;
		
		if (DEBUG) {
			System.out.println("executeLocal()");
		}
		
		if (!this.isPositionValid(rowTo, colTo)) {
			throw new IllegalPositionException();
		}
		
		//if the requested position is empty 
		if (board[rowTo][colTo] == NO_PIECE) {
			moveStatus = VALID_UNFINISHED;
			if (chains)
				System.out.println("ball unmarked = " + Integer.toHexString(ball));
			//MARK the ball
			ball |= MARKER_MASK;
			this.numBallsMarked ++;
			if (chains)
				System.out.println("ball marked = " + Integer.toHexString(ball));
			//put the ball into it's position
			this.board[rowTo][colTo] = ball;
			if (chains)
				System.out.println("ball at specified position = " + Integer.toHexString(this.board[rowTo][colTo]));
			this.numBallsOnBoard ++;
			if (chains)
				System.out.println("numBallsOnBoard = " + this.numBallsOnBoard);
			//mark all balls chained to the new ball that have the same color
			this.markBallsConnectedTo(rowTo, colTo, true);
			if (chains)
				System.out.println("numBallsMarked = " + this.numBallsMarked);
			//if 3 or more balls with the same color are connected disappear them
			if (this.numBallsMarked >= 3) {
				moveStatus = VALID_FINISHED;
				if (chains)
					System.out.println("deleting all marked balls");
				//delete Marked Balls
				this.deleteBalls(MARKER_MASK, true);
				if (chains)
					System.out.println("looking for pieces to drop on oponent");
				//since there was a structural board change look for pieces that are no longer connected
				//to the top and drop them on the oponent :)
				//first mark all balls in any way chained to the top row
				for (byte i = 0; i < this.board[0].length; i++) {
					if (board[0][i] != NO_PIECE) {
						//mark the current ball
						board[0][i] |= MARKER_MASK;
						//and all balls cunnected to it
						this.markBallsConnectedTo((byte) 0, i, false);
					}
				}
				//then get all unmarked balls off the board and into the play
				moveUnmarkedBallsToPlay();
				if (chains || DEBUG)
					System.out.println("here are the bombs:\n" + this.currentPlay.toString());
			}
			if (DEBUG)
				System.out.println("resetting the STATE - unmarking all balls & clearing counters");
			//unmarkBalls
			this.unmarkAllBalls();
			//reset the maked ball counter
			this.numBallsMarked = 0;
		}
		else {
			//else not valid move
			if (DEBUG)
				System.out.println("There is already a piece on that position!");
			moveStatus = INVALID;
		}

		//if this ball is to go into the last row
		//the local player looses - unless this move drops balls
		if ( (rowTo == BOARD_ROWS - 1) && ( moveStatus != VALID_FINISHED ) ) {
			if (DEBUG)
				System.out.println("returning LOSE");
			return LOSE;
		}
		if (DEBUG)
			System.out.println("returning " + moveStatus);
		return moveStatus;
	}
	
	private byte determineUniqueBall(byte row, byte col) {
		if (DEBUG)
			System.out.println("determineRemoteBall()");
		byte remoteBall = BALL_BLUE;
		boolean[] colors = new boolean[NUM_BALL_TYPES];
		byte[][] touchingBalls = this.getTouchingBalls(row, col, false, false);
		for (byte i = 0; i < touchingBalls.length; i++) {
			try {
				if (touchingBalls[i] != null) {
					byte ball = this.getPieceAtPosition(touchingBalls[i][0], touchingBalls[i][1]);
					byte color = (byte) (ball & COLOR_MASK);
					//colors begin with 1 however array index begins with 0, this adjusts
					colors[color -1] = true;
				}
			} catch (IllegalPositionException e) {
				//This should NEVER happen
				e.printStackTrace();
			}
		}
		while (true) {
			if (DEBUG)
				System.out.println("REMOTE BALL");
			remoteBall = this.getRandomBall();
			if (!colors[remoteBall -1]) {
				if (DEBUG)
					System.out.println("Remote ball determined to be: " + remoteBall);
				return remoteBall;
			}
		}
	}
	
	private void moveUnmarkedBallsToPlay() {
		for (byte i = 0; i < board.length; i++) {
			for (byte j = 0; j < board[i].length; j++) {
				byte curBall = board[i][j];
				if (curBall != NO_PIECE) {
					if ((curBall & MARKER_MASK) == 0x0 ) {
						this.numBallsSent++;
						//add a remote move for the oponent
						this.currentPlay.addMove(REMOTE_MOVE, curBall, i, j);
						//remove this ball from local board 
						board[i][j] = NO_PIECE;
					}
				}
			}
		}		
	}

	private void unmarkAllBalls() {
		for (byte i = 0; i < board.length; i++) {
			for (byte j = 0; j < board[i].length; j++) {
				byte curBall = board[i][j];
				if (curBall != NO_PIECE) {
					board[i][j] = (byte) (curBall & ~MARKER_MASK);
				}
			}
		}		
	}
	
	private void deleteBalls(byte mask, boolean set) {
		for (byte i = 0; i < board.length; i++) {
			for (byte j = 0; j < board[i].length; j++) {
				if (board[i][j] != NO_PIECE) {
					byte result = (byte) (board[i][j] & mask);
					//mask MUST be set
					if (set) {
						if (result == mask) {
							this.board[i][j] = NO_PIECE;
							this.numBallsOnBoard --;
						}
					}
					//mask MUST NOT be set
					else {
						if (result == 0x0) {
							this.board[i][j] = NO_PIECE;
							this.numBallsOnBoard --;
						}
					}
				}
			}
		}
	}

	private void markBallsConnectedTo(byte row, byte col, boolean sameColorOnly) {
		if (chains)
			System.out.println("\tmarking balls connected with "+row+", "+col+"  sameColorOnly = " + sameColorOnly);
		//i want balls that are touching this one and have the same color and are unmarked
		byte[][] touchingBalls = getTouchingBalls(row, col, sameColorOnly, true);
		for (byte i = 0; i < touchingBalls.length; i++) {
			if (touchingBalls[i] != null && touchingBalls[i].length != 0)
				markBallsConnectedTo(touchingBalls[i][0], touchingBalls[i][1], sameColorOnly);
		}
	}
	
	private byte[][] getTouchingBalls(byte row, byte col, boolean sameColorOnly, boolean mark) {
		byte ball = NO_PIECE;
		try {
			ball = this.getPieceAtPosition(row, col);
		} catch (IllegalPositionException e) {
			if (DEBUG)
				e.printStackTrace();
		}
		if (chains)
			System.out.println("\t\tgetting touching balls for ball  " +row+", "+col+"  color = " + Integer.toHexString(ball));
		//maximum number of balls that can touch one ball = 6
		byte[][] touchingBalls = new byte[6][];
		byte next = 0;
		
		//The balls immediately to the left and right are calculated the same regardless of row even or odd

		//look at the ball to the left only if not on the left-most ball
		if (col > 0) {
			if (isATouchingBall(ball, row, (byte) (col - 1), sameColorOnly))
				touchBall(touchingBalls, next++, row, (byte) (col - 1), mark);
		}

		//look at the ball to the right only if not on the right-most ball
		if (col < board[row].length -1)
		if (isATouchingBall(ball, row, (byte) (col + 1), sameColorOnly))
			touchBall(touchingBalls, next++, row, (byte) (col + 1), mark);
		
		//if the ball is in a row with SHORT_COLS, calculate the touching this way:
		if (board[row].length == SHORT_COLS) {
			if (chains)
				System.out.println("\t\tSHORT_COLS row");
			//balls below only if not on the last row
			if (row < board.length - 1) {
				if (isATouchingBall(ball, (byte) (row + 1), col, sameColorOnly)) 
					touchBall(touchingBalls, next++, (byte) (row + 1), col, mark);
				if (isATouchingBall(ball, (byte) (row + 1), (byte) (col + 1), sameColorOnly))
					touchBall(touchingBalls, next++, (byte) (row + 1), (byte) (col + 1), mark);
			}
			//balls above - only if row > 0
			if (row > 0) {
				if (isATouchingBall(ball, (byte) (row - 1), col, sameColorOnly))
					touchBall(touchingBalls, next++, (byte) (row - 1), col, mark);
				if (isATouchingBall(ball, (byte) (row - 1), (byte) (col + 1), sameColorOnly)) 
					touchBall(touchingBalls, next++, (byte) (row - 1), (byte) (col + 1), mark);
			}

		}
		//else it must be in a row with FULL_COLS so calculate the touching this way:
		else {
			if (chains)
				System.out.println("\t\tFULL_COLS row");
			//balls below only if not on the last row
			if (row < board.length - 1) {
				if (isATouchingBall(ball, (byte) (row + 1), (byte) (col - 1), sameColorOnly)) 
					touchBall(touchingBalls, next++, (byte) (row + 1), (byte) (col - 1), mark);
				if (isATouchingBall(ball, (byte) (row + 1), col, sameColorOnly)) 
					touchBall(touchingBalls, next++, (byte) (row + 1), col, mark);
			}
			//balls above - only if row > 0
			if (row > 0) {
				if (isATouchingBall(ball, (byte) (row - 1), (byte) (col - 1), sameColorOnly)) 
					touchBall(touchingBalls, next++, (byte) (row - 1), (byte) (col - 1), mark);
				if (isATouchingBall(ball, (byte) (row - 1), col, sameColorOnly)) 
					touchBall(touchingBalls, next++, (byte) (row - 1), col, mark);
			}
		}
		return touchingBalls;
	}

	private boolean isATouchingBall(byte ball, byte row, byte col, boolean sameColorOnly) {
		byte touchingBall;
		if (chains)
			System.out.println("\t\t\tisATouchingBall() looking @ ball  " +row+", "+col+"  color = " + Integer.toHexString(ball));
		if (isPositionValid(row, col)) {
			touchingBall = board[row][col];
			//if there is a piece and it isn't marked already
			if ( touchingBall != NO_PIECE && ((touchingBall & MARKER_MASK) != MARKER_MASK) )  {
				if (sameColorOnly) {
					if ( ! ( (touchingBall & COLOR_MASK) == (ball & COLOR_MASK) ) ) {
						if (chains)
							System.out.println("\t\t\treturning FALSE - not same color");
						return false; 
					}
				}
				//if all is fine
				if (chains)
					System.out.println("\t\t\treturning TRUE");
				return true;
			}
			//else there id no piece OR marked piece
			if (chains)
				System.out.println("\t\t\teither NO_PIECE or MARKED_PIECE");
		}
		if (chains)
			System.out.println("\t\t\treturning FALSE");
		return false;
	}

	private void touchBall(byte[][] touchingBalls, byte pos, byte row, byte col, boolean mark) {
		touchingBalls[pos] = new byte[2];
		touchingBalls[pos][0] = row;
		touchingBalls[pos][1] = col;
		//mark the ball
		if (mark) {
			this.board[row][col] = (byte) (this.board[row][col] | MARKER_MASK);
			this.numBallsMarked ++;
		}
	}
	

	public byte getCurrentBall() {
		return this.currentBall;
	}
	
	public byte getNextBall() {
		return this.nextBall;
	}

	private byte getRandomBall() {
		byte ball = (byte) (Math.abs(this.random.nextInt() % NUM_BALL_TYPES) + 1);
		if (DEBUG)
			System.out.println("--- RETURNING RANDOM BALL VALUE : " + ball);
		return ball;
	}

	public byte updateCurrentAndNextBall() {
		byte tmp = this.currentBall;
		this.currentBall = this.nextBall;
		this.nextBall = this.getRandomBall();
		return tmp;
	}
	
/////////////////////////////////////////////////////////////////////////////////////////////
	
	public Play getPlay() {
		return this.currentPlay;
	}

	public byte getPieceAtPosition(byte row, byte col) throws IllegalPositionException {
		if (!isPositionValid(row, col)) {
			if (DEBUG)
				throw new IllegalPositionException("can't get piece for row: " + row + " col: " + col);
			throw new IllegalPositionException();
		}
		return board[row][col];
	}

	public boolean isPositionValid(byte row, byte col) {
		if (row >= 0 && row < BOARD_ROWS) {
			if (col >= 0 && col < board[row].length) {
				return true;
			}
		}
		return false;
	}

	public byte[][] getBoard() {
		return this.board;
	}

	public void setBoard(byte[][] board) {
		this.board = board;
	}

	public short getNumBallsRecieved() {
		return this.numBallsRecieved;
	}
	public short getNumBallsSent() {
		return this.numBallsSent;
	}
	
	
	public void clearReferences() {
	}

}
