package com.cotopia.games;

import javax.microedition.lcdui.*;


/**
 * @author karel herink
 */
public class Ball {

	private final static boolean DEBUG = false;
	
	//move UP (posY-1) one pixel every ballPeriodMS
	private static int ballPeriodMS = 20;
	private byte type;
	
	//the region in which the ball moves
	private int top;
	private int left;
	private int right;

	//the position and size of the ball
	private volatile int startPosX;
	private volatile int startPosY;
	private volatile int posX;
	private volatile int posY;
	private int ballSize;
	private int radius;
	private int pixelsPerUnit;

	//the direction of the ball is controlled by these variables
	private int deltaX;
	private int deltaY;
	private int b;
	
	private long ballLastRepaintTime;
	private int anchor = Graphics.VCENTER | Graphics.HCENTER;
	
	private BallShotBoard board;

	
	private int[] exitPoint = new int[2];
	private static int counter = 0;
	private int centerX;
	private int centerY;
	
	
	/**
	 * Constructor defines the region in which the ball moves as well
	 * as its starting position.
	 */
	Ball(byte type, BallShotBoard board, Gun cannon, short centerX, short centerY, short startPosX, short startPosY) {
		counter++;
		this.type = type;
		this.board = board;
		this.centerX = centerX;
		this.centerY = centerY;
		this.startPosX = startPosX;
		this.startPosY = startPosY;
		this.posX = startPosX;
		this.posY = startPosY;
		this.ballSize = board.fieldEdgeLenght;
		this.radius = this.ballSize / 2;
		this.pixelsPerUnit = (this.radius / 3) *2;
		this.deltaX = this.startPosX - centerX;
		this.deltaY = this.startPosY - centerY;
		this.b = this.linearCalculateB(this.deltaX, this.deltaY, this.posX, this.posY);
		this.left = board.leftXBoundry;
		this.right = board.rightXBoundry;
		this.top  = board.topYBoundry;
		this.ballLastRepaintTime = System.currentTimeMillis();
		
		if (DEBUG) {
			System.out.println(this.toString());
		}
	}
 
	public void updateBallPosition() {
		
		int lastPosX = this.posX;
//		int lastPosY = this.posY;
		
		long now = System.currentTimeMillis();
		long diff = now - this.ballLastRepaintTime;
		int units = (int) (diff / ballPeriodMS);

		//THIS IS VERY IMPORTANT HERE :)  -- took 24 hours of debugging!! 
		if (units == 0)
			return;
		
		units *= this.pixelsPerUnit; 
		
		//this ensures that we dont skip more than the ball diameter in the Y direction
		//that could cause us to skip over balls
		units = units > this.ballSize ? ballSize : units;
		
		if (DEBUG)
			System.out.println("\n\nupdateBallPosition() Y = " + posY + " UP BY " + units);		
		
		//move up units (our speed is measurred only upwards) and since the topY = 0 we have to
		//subtract to go up
		this.posY = posY - units;
		this.posX = b == 0 ? this.posX : this.linearSolveX(this.deltaX, this.deltaY, posY, this.b);
		
		//this ensures that we dont skip more than the ball radius in the X direction
		//that could cause us to skip over balls
		int xDiff = this.posX - lastPosX;
		if (Math.abs(xDiff) > this.radius) {
			this.posX = xDiff > 0 ? lastPosX + this.radius : lastPosX - this.radius;
			this.posY = this.linearSolveY(this.deltaX, this.deltaY, this.posX, this.b);
		}
		
		if (DEBUG)
			System.out.println("X = " + posX + "\tY = " + posY);
		
		this.compensateLeftWallCollision();
		this.compensateRightWallCollision();
		
		//if above the top OR the new position is in collision with a ball
		if (isBallCollision() || isTopCollision())
			this.board.setBall(this.type, posX, posY);
		
		this.ballLastRepaintTime = now;
		
		if (DEBUG)
			System.out.println("BALL X = " + posX + "\tY = " + posY);
		
		//TODO : repaint the ball only
		// board.repaint();//(posX - 1, posY - 1, ballSize + 2, ballSize + 2);
		// assign the new position and queue a repaint
	}
	
	private boolean isBallCollision() {
		boolean ballWillBeSet = false;
		short cornerFraction = (short) ((this.radius / 3) * 2);
		if (	//top
				board.isBallCollisionAtPosition(posX, posY - this.radius) || 
				//bottom
				board.isBallCollisionAtPosition(posX, posY + this.radius) ||
				//left
				board.isBallCollisionAtPosition(posX - (this.radius / 5) * 4 , posY) ||
				//right
				board.isBallCollisionAtPosition(posX + (this.radius / 5) * 4, posY) ||
				//top-right
				board.isBallCollisionAtPosition(posX + cornerFraction, posY - cornerFraction) ||
				//top-left
				board.isBallCollisionAtPosition(posX - cornerFraction, posY - cornerFraction) ||
				//bottom-right
				board.isBallCollisionAtPosition(posX + cornerFraction, posY + cornerFraction) ||
				//bottom-left
				board.isBallCollisionAtPosition(posX - cornerFraction, posY + cornerFraction)
		)
		{
			if (DEBUG)
				System.out.println("collision with a ball");
			ballWillBeSet = true;
			//it will be better to move along more sensitive axis
			if ( Math.abs(this.deltaX) > Math.abs(this.deltaY)) {
				if (DEBUG)
					System.out.println("moving along X-axis");
				
				while (board.isBallCollisionAtPosition(posX, posY)) {
					posX = this.deltaX > 0 ? posX-1 : posX+1;
					if (this.deltaY != 0) //under current coditions deltaY will never be 0
						posY = this.linearSolveY(this.deltaX, this.deltaY, posX, this.b);
					if (DEBUG)
						System.out.println("\tX = " + posX + "\tY = " + posY);
				}
			}
			else {
				if (DEBUG)
					System.out.println("moving along Y-axis");
				do {
					posY = posY + 1;
					if (this.deltaX != 0)
						posX = this.linearSolveX(this.deltaX, this.deltaY, posY, this.b);
					if (DEBUG)
						System.out.println("\tX = " + posX + "\tY = " + posY);
				} while (board.isBallCollisionAtPosition(posX, posY));
			}
		}
		return ballWillBeSet;
	}
	
	
	private void compensateLeftWallCollision() {
		if ((posX - this.radius) <= this.left) {
			if (DEBUG)
				System.out.println("collision with the left wall,  (posX - this.radius) = " + (posX - this.radius));			
			
			this.exitPoint[0] = posX;
			this.exitPoint[1] = posY;
			
			if ( Math.abs(this.deltaX) > Math.abs(this.deltaY)) {
				if (DEBUG)
					System.out.println("\n\tmoving along X-axis");
				do {
					posX = this.deltaX > 0 ? posX - 1 : posX + 1; 
					posY = this.linearSolveY(this.deltaX, this.deltaY, posX, this.b);
					if (DEBUG && posX > -99)
						System.out.println("\n\tX = " + posX + "\tY = " + posY);
				} while ((posX - this.radius) <= this.left);
			}
			else {
				if (DEBUG)
					System.out.println("\n\tmoving along Y-axis");
				do {
					posY = posY + 1;
					posX = this.linearSolveX(this.deltaX, this.deltaY, posY, this.b);
					if (DEBUG && posX > -99)
						System.out.println("\n\tX = " + posX + "\tY = " + posY);
				} while ((posX - this.radius) <= this.left);
			}
			
			if (DEBUG)
				System.out.println("\n\tSwitched DX when x= " + this.posX + " , y = " + this.posY);
			
			//switch slope
			this.deltaX = -deltaX;
			//recalculate 'b'
			this.b = this.linearCalculateB(this.deltaX, this.deltaY, posX, posY);
		}
	}
	
	private void compensateRightWallCollision() {
		if ((posX + this.radius ) >= this.right) {
			if (DEBUG)
				System.out.println("collision with the right wall");
			if ( Math.abs(this.deltaX) > Math.abs(this.deltaY)) {
				if (DEBUG)
					System.out.println("moving along X-axis");
				do {
					posX = this.deltaX > 0 ? --posX : ++posX; 
					posY = this.linearSolveY(this.deltaX, this.deltaY, posX, this.b);
					if (DEBUG)
						System.out.println("\tX = " + posX + "\tY = " + posY);
				} while ((posX + this.radius ) >= this.right);
			}
			else {
				if (DEBUG)
					System.out.println("moving along Y-axis");
				do {
					posY = posY + 1;
					posX = this.linearSolveX(this.deltaX, this.deltaY, posY, this.b);
					if (DEBUG)
						System.out.println("\tX = " + posX + "\tY = " + posY);
				} while ((posX + this.radius ) >= this.right -1);
			}
			//switch slope
			this.deltaX = deltaX * (-1);
			//recalculate 'b'
			this.b = this.linearCalculateB(this.deltaX, this.deltaY, posX, posY);
		}
	}
	
	private boolean isTopCollision() {
		boolean ballWillBeSet = false;
		if ((posY - this.radius) < this.top) {
			if (DEBUG)
				System.out.println("above the top");
			ballWillBeSet = true;
			//it will be better to move along more sensitive axis
			if ( Math.abs(this.deltaX) > Math.abs(this.deltaY)) {
				if (DEBUG)
					System.out.println("moving along X-axis");
				do {
					posX = this.deltaX > 0 ? posX-1 : posX+1;
					posY = this.linearSolveY(this.deltaX, this.deltaY, posX, this.b);
					if (DEBUG)
						System.out.println("\tX = " + posX + "\tY = " + posY);
				} while ((posY - this.radius) < this.top);
			}
			else {
				if (DEBUG)
					System.out.println("moving along Y-axis");
				do {
					posY = posY + 1;
					if (this.deltaX != 0)
						posX = this.linearSolveX(this.deltaX, this.deltaY, posY, this.b);
					if (DEBUG)
						System.out.println("\tX = " + posX + "\tY = " + posY);
				} while ((posY - this.radius) < this.top);
			}
		}
		return ballWillBeSet;
	}
	
	
	
	
    /**
     * Paint the ball.
     */
    void paint(Graphics g) {
    	if (this.board.running) {
	    	this.updateBallPosition();
	    	g.setColor(0);
	    	//the condition may have changed in updateBallPosition()
	    	if (this.board.running)
	    		g.drawImage((Image)this.board.getImageCreator().getImage(this.type, BallShotBoard.UNKNOWN), posX, posY, this.anchor);
    	}
    }

    boolean inside(int x1, int y1, int x2, int y2) {
        return (posX <= x2) 
            && (posY <= y2)
            && ((posX + ballSize) >= x1) 
            && ((posY + ballSize) >= y1);
    }


    // controls the speed of all balls; delay in centiseconds
    static void slower() {
    	ballPeriodMS += 20;
    	if (ballPeriodMS > 200) 
    		ballPeriodMS = 200;
    }
    static void faster() {
    	ballPeriodMS -= 20;
    	if (ballPeriodMS < 0) 
    		ballPeriodMS = 0;
    }

    //to calculate 'b' we need slope and a point : b = y - mx 
    public int linearCalculateB(int deltaX, int deltaY, int pointX, int pointY) {
    	if (deltaX == 0)
    		return 0;
    	return pointY - ((deltaY * pointX) / deltaX);
    }
    
    public int linearSolveX(int deltaX, int deltaY, int y, int b) {
    	return  ((y - b) * deltaX) / deltaY;
    }

    public int linearSolveY(int deltaX, int deltaY, int x, int b) {
    	return (deltaY * x) / deltaX + b;
    }
    
    public String toString() {
    	if (DEBUG) {
    		return 
    		"\nID = " + counter
    		+ "\nRADIUS = " + this.radius
    		+ "\nLEFT  = " + this.left + "\nRIGHT = " + this.right + "\nTOP   = " + this.top
    		+ "\ncenterX = " + this.centerX
    		+ "\ncenterY = " + this.centerY
    		+ "\nstartX = " + this.startPosX
    		+ "\nstartY = " + this.startPosY
    		+ "\nDX = " + this.deltaX + "\nDY = " + this.deltaY + "\nB = " + this.b 
    		+ "\nEXITED @ x = " + this.exitPoint[0] + ", y = " + this.exitPoint[1]
    		+ "\nX = " + posX
    		+ "\nY = " + posY;
    	}
    	return super.toString();
    }
    
}
