package com.cotopia.games;

import javax.microedition.lcdui.Graphics;
import javax.microedition.lcdui.Image;

import com.cotopia.games.BallShotRules;

/**
 * @author karel herink
 */
public class Gun {

	private BallShotBoard board;
	private BallShotRules rules;
	private boolean paused;

	short cannonTipX;
	short cannonTipY;
	
	short canonShotCenterX;
	short canonShotCenterY;
	
	private short cannonTipXBoundryLeft;
	private short cannonTipXBoundryRight;
	
	//after how long of a period does the canon change position
	short canonPeriodMS;
	
	
	//by how much in the X direction does the position change
	//is initialized to vary along with the fieldEdgeLenght
	private short canonDeltaXSpeed;
	private long canonLastRepaintTime;
	private byte ballRadius;
	
	protected Gun(BallShotBoard board, byte ballDiameter, int ballFireLimitMS) {
		this.ballRadius = (byte) (ballDiameter / 2);
		this.board = board;
		this.rules = board.getRules();
		
		this.canonShotCenterX = (short) (board.canonLeft + (board.canonWidth / 2));
		this.canonShotCenterY = (short) (board.canonTop + board.canonHeight);
		
		this.cannonTipX = this.canonShotCenterX; //the canon starts nd moves towards the right boundry
		this.cannonTipY = (short) (board.ballsHeight + board.ballsOffsetY);
		
		this.cannonTipXBoundryLeft = (short) (board.ballsOffsetX + board.fieldEdgeLenght / 2);
		this.cannonTipXBoundryRight = (short) (board.ballsOffsetX + board.ballsWidth - board.fieldEdgeLenght / 2);
		
		this.canonPeriodMS = 120;
		this.canonDeltaXSpeed = (short) (board.fieldEdgeLenght / 5);
	}

	void paint(Graphics g) {
		if (this.board.running) {
			g.setColor(0xffffff);
			g.fillRect(board.canonLeft, board.canonTop, board.canonWidth, board.canonHeight);
			calculateCanonTip();
			g.setColor(0xbb00ff);
			g.drawLine(this.canonShotCenterX-2, this.canonShotCenterY, this.cannonTipX-2, this.cannonTipY);
			g.setColor(0xf85ecc);
			g.drawLine(this.canonShotCenterX-1, this.canonShotCenterY, this.cannonTipX-1, this.cannonTipY);
			g.setColor(0xff0000);
			g.drawLine(this.canonShotCenterX, this.canonShotCenterY, this.cannonTipX, this.cannonTipY);
			g.setColor(0xf85ecc);
			g.drawLine(this.canonShotCenterX+1, this.canonShotCenterY, this.cannonTipX+1, this.cannonTipY);
			g.setColor(0xbb00ff);
			g.drawLine(this.canonShotCenterX+2, this.canonShotCenterY, this.cannonTipX+2, this.cannonTipY);
			g.setColor(0x000000);
			//g.setColor(0xddffcc);
			//g.fillArc(this.canonShotCenterX - 3*this.ballRadius, this.canonShotCenterY - 3*this.ballRadius, this.ballRadius*6, this.ballRadius*6, 0, 180);
			//g.setColor(0x000000);
			//g.drawArc(this.canonShotCenterX - 3*this.ballRadius, this.canonShotCenterY - 3*this.ballRadius, this.ballRadius*6, this.ballRadius*6, 0, 180);
			g.drawImage((Image) board.getImageCreator().getImage(rules.getCurrentBall(), BallShotBoard.UNKNOWN), 
					this.canonShotCenterX, this.canonShotCenterY, Graphics.BOTTOM | Graphics.HCENTER);
			Image canonTile = ((BallShotImageCreator)board.getImageCreator()).getIcon(BallShotImageCreator.GUN_FOREGROUND);
			if (canonTile != null)
				g.drawImage(canonTile, board.canonLeft, board.canonTop, Graphics.TOP | Graphics.LEFT);
			g.drawRect(board.canonLeft, board.canonTop, board.canonWidth, board.canonHeight);
		}
	}

	private void calculateCanonTip() {
		long now = System.currentTimeMillis();
		if (this.paused) {
			this.canonLastRepaintTime = now;
			return;
		}
		long repaintDiff = now - this.canonLastRepaintTime;
		short units = (short) (repaintDiff / canonPeriodMS);
		if (units != 0) {
			this.canonLastRepaintTime = now;
			this.cannonTipX += (short) (units * this.canonDeltaXSpeed);
			if (this.cannonTipX >= this.cannonTipXBoundryRight) {
				//if past maximum range then set to maximum range
				this.cannonTipX = this.cannonTipXBoundryRight;
				//start movin in the oposite direction
				this.canonDeltaXSpeed = (short) -this.canonDeltaXSpeed;
			}
			else if (this.cannonTipX <= this.cannonTipXBoundryLeft) {
				//if past maximum range then set to maximum range
				this.cannonTipX = this.cannonTipXBoundryLeft;
				//start movin in the oposite direction
				this.canonDeltaXSpeed = (short) -this.canonDeltaXSpeed;
			}
		}
	}

	public void shoot() {
		if (this.board.getBall() != null)
			return;
		this.calculateCanonTip();

		Ball ball = new Ball(rules.updateCurrentAndNextBall(), board, this, this.canonShotCenterX, this.canonShotCenterY, this.cannonTipX, this.cannonTipY);
		this.board.setBall(ball);
		//this.paused = true;
	}

	public short faster() {
		return ((this.canonPeriodMS -= 10) < 30 ? this.canonPeriodMS = 30 : this.canonPeriodMS);
	}

	public short slower() {
		return ((this.canonPeriodMS += 10) > 120 ? this.canonPeriodMS = 120 : this.canonPeriodMS);
	}
	
	public void setPaused(boolean paused) {
		this.paused = paused; 
	}	

	public boolean isPaused() {
		return this.paused;
	}
	
	public void setRight() {
		this.canonDeltaXSpeed = (short) Math.abs(this.canonDeltaXSpeed);
	}

	public void setLeft() {
		this.canonDeltaXSpeed = (short) - Math.abs(this.canonDeltaXSpeed);
	}
	
}
